/**
 *    FrameMouseActions.java
 *    Copyright (C) 2010 New Zealand Digital Library, http://expeditee.org
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.expeditee.gui;

import java.awt.event.MouseEvent;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.expeditee.actions.Actions;
import org.expeditee.actions.Misc;
import org.expeditee.actions.Navigation;
import org.expeditee.core.AxisAlignedBoxBounds;
import org.expeditee.core.Colour;
import org.expeditee.core.OSManager;
import org.expeditee.core.Point;
import org.expeditee.core.swing.SwingConversions;
import org.expeditee.gio.gesture.Gesture;
import org.expeditee.gio.gesture.GestureListener;
import org.expeditee.gui.indirect.mouse.IndirectMouseActions;
import org.expeditee.gui.indirect.mouse.MouseAction;
import org.expeditee.gui.indirect.mouse.MouseInfo;
import org.expeditee.gui.Frame;
import org.expeditee.io.ExpClipReader;
import org.expeditee.io.ItemSelection;
import org.expeditee.io.ItemSelection.ExpDataHandler;
import org.expeditee.items.Circle;
import org.expeditee.items.Constraint;
import org.expeditee.items.Dot;
import org.expeditee.items.Item;
import org.expeditee.items.Item.HighlightMode;
import org.expeditee.items.ItemAppearence;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Line;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;
import org.expeditee.items.UserAppliedPermission;
import org.expeditee.items.XRayable;
/*import org.expeditee.items.widgets.ButtonWidget; TODO: Reinstate. cts16
import org.expeditee.items.widgets.InteractiveWidget;
import org.expeditee.items.widgets.WidgetCorner;
import org.expeditee.items.widgets.WidgetEdge;*/
import org.expeditee.settings.UserSettings;
import org.expeditee.settings.experimental.ExperimentalFeatures;
import org.expeditee.stats.SessionStats;

public class FrameMouseActions
{

	private static int _lastMouseClickModifiers = 0;

	private static MouseEvent _lastMouseDragged;

	private boolean _autoStamp = false;
	
	private FrameMouseActions() {
		IndirectMouseActions.getInstance().setBackAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				//if user is not pointing at something, this is a back
				if (info.isControlDown || info.isShiftDown)
					forward();
				else
					back();
				return null;
			}
		});
		IndirectMouseActions.getInstance().setTDFCAction(new MouseAction() {
			@Override
			public List<Item> exec(final MouseInfo info) {
				// check for TDFC permission
				if (!info.clicked.hasPermission(UserAppliedPermission.createFrames)) {
					MessageBay.displayMessage("Insufficient permission to TDFC (Top Down Frame Creation) from that item");
					return null;
				}

				if (info.clicked.isOldTag())
					return null;

				try {
					tdfc(info.clicked);
				} catch (RuntimeException e) {
					e.printStackTrace();
					MessageBay.errorMessage("Top Down Frame Creation (TDFC) error: " + e.getMessage());
				}
				return null;
			}
		});
		IndirectMouseActions.getInstance().setRespondToFrameNameClickAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				if (info.isControlDown)
					Navigation.PreviousFrame(false);
				else
					Navigation.NextFrame(false);
				return null;
			} 		
		});
		IndirectMouseActions.getInstance().setFollowLinkAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				/*
				 * Dont save the frame if we are moving to an old version of
				 * this frame because everytime we save with the old tag... the
				 * frame is backed up
				 */
				if (!info.clicked.isOldTag())
					FrameIO.SaveFrame(DisplayIO.getCurrentFrame());

				Navigation.setLastNavigationItem(info.clicked);
				load(info.clicked.getAbsoluteLink(), info.clicked.getLinkHistory());
				// DisplayIO.UpdateTitle();
				return null;
			}
		});
		IndirectMouseActions.getInstance().setExecuteActionAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				info.clicked.performActions();
				info.clicked.setHighlightMode(HighlightMode.None);
				getInstance().refreshHighlights();
				return null;
			} 
		});
		IndirectMouseActions.getInstance().setCreateRectangleAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				final List<Item> copies = new ArrayList<Item>();
				Item[] d = new Item[RECTANGLE_CORNERS];
				// create dots
				Frame current = DisplayIO.getCurrentFrame();
				for (int i = 0; i < d.length; i++) {
					d[i] = current.createDot();
					copies.add(d[i]);
				}

				current.nextDot();

				// create lines
				copies.add(new Line(d[0], d[1], current.getNextItemID()));
				copies.add(new Line(d[1], d[2], current.getNextItemID()));
				copies.add(new Line(d[2], d[3], current.getNextItemID()));
				copies.add(new Line(d[3], d[0], current.getNextItemID()));

				new Constraint(d[0], d[1], current.getNextItemID(), Constraint.HORIZONTAL);
				new Constraint(d[2], d[3], current.getNextItemID(), Constraint.HORIZONTAL);
				new Constraint(d[1], d[2], current.getNextItemID(), Constraint.VERTICAL);
				new Constraint(d[3], d[0], current.getNextItemID(), Constraint.VERTICAL);

				anchor(new ArrayList<Item>(copies));
				pickup(d[3]);
				d[3].setHighlightMode(HighlightMode.Normal);

				SessionStats.CreatedItems(copies);
				copies.clear();
				return copies;
			} 
		});
		IndirectMouseActions.getInstance().setExtendLineAction(new MouseAction() {
			@Override
			public List<Item> exec(final MouseInfo info) {
				Line onLine = (Line) info.clicked;
				Line newLine = onLine.copy();
				Item end = newLine.getEndItem();
				Item start = newLine.getStartItem();
				end.setPosition(MouseX, MouseY);
				start.setPosition(MouseX, MouseY);
				onLine.autoArrowheadLength();
				// anchor the start
				anchor(start);
				// attach the line to the cursor
				pickup(end);

				List<Item> toMerge = new LinkedList<Item>();
				toMerge.add(newLine.getStartItem());
				toMerge.add(newLine);

				// Make sure the highlighting is shown when the end is
				// anchored
				end.setHighlightMode(Item.HighlightMode.Normal);
				merge(toMerge, info.clicked);
				// anchor(left);
				// FreeItems.getInstance().clear();
				FrameGraphics.Repaint();
				updateCursor();
				return null;
			} 
		});
		IndirectMouseActions.getInstance().setMakeGroupCopyAction(new MouseAction() {
			@Override
			public List<Item> exec(final MouseInfo info) {
				List<Item> copies = new ArrayList<Item>();
				// Set the selection mode for the items that were clicked in
				Collection<Item> enclosed = getFullyEnclosedItems(info.clickedIn);
				if (enclosed.size() == 0) {
					MessageBay.displayMessage("Insufficient permission to copy items");
				} else {
					copies = copy(enclosed);
					clearParent(copies);
					ItemUtils.Justify(copies);
					pickup(copies);
					for (Item i : info.clickedIn) {
						i.setHighlightMode(HighlightMode.None);
					}
				}
				return copies;
			}
		});
		IndirectMouseActions.getInstance().setMakeCopyAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				List<Item> copies = ItemUtils.UnreelLine(info.clicked, _controlDown);
				// Copies will NOT be null if the user right clicked on a point
				if (copies == null) {
					Collection<Item> originals = info.clicked.getConnected();
					copies = ItemUtils.CopyItems(originals, _extrude);
					// if this is the title of the frame, link it to the frame
					if (originals.size() == 1 && copies.size() == 1) {
						Item copy = copies.get(0);
						Item original = originals.iterator().next();
						if (original.getLink() == null
								&& original.isFrameTitle()) {
							// save the frame after copying
							// i.getParent().setChanged(true);
							copy.setLink(original.getParentOrCurrentFrame()
									.getName());
						}
					}

					FrameGraphics.changeHighlightMode(info.clicked,
							HighlightMode.None);

					if (!_extrude)
						clearParent(copies);
				}
				
				ItemUtils.Justify(copies);

				pickup(copies);
				return copies;
			}	
		});
		IndirectMouseActions.getInstance().setExtrudeAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				List<Item> copies = null;
				List<Item> originals = new ArrayList<Item>();
				// remove any lines that dont have both endpoints
				// floating
				for (Item i : FreeItems.getInstance()) {
					if (i.isFloating())
						originals.add(i);
				}
				if (copies == null)
					copies = ItemUtils.CopyItems(originals, _extrude);
				for (Item i : copies)
					i.setOffset(0, 0);
				anchor(FreeItems.getInstance());
				// Move isnt working right for extruding!!
				// move(copies);
				pickup(copies);
				return copies;
			} 	
		});
		IndirectMouseActions.getInstance().setRubberBandingCopyAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				List<Item> copies = new ArrayList<Item>();
				if (info.clicked != null) {
					Collection<Item> leftOver = merge(FreeItems.getInstance(), info.clicked);
					anchor(leftOver);
				}
				// This is executed when the user is putting down a line
				// endpoint and unreeling. ie. Normal unreeling
				copies = ItemUtils.UnreelLine(FreeItems.getInstance(), _controlDown);

				if (copies == null) copies = copy(FreeItems.getInstance());
				
				anchor(FreeItems.getInstance());
				for (Item i : copies) i.setOffset(0, 0);
				// need to move to prevent cursor dislocation
				move(copies);
				pickup(copies);
				return copies;
			}		
		});
		IndirectMouseActions.getInstance().setRubberBandingCornerAction(new MouseAction() {
			@Override
			public List<Item> exec(final MouseInfo info) {
				List<Item> copies = new ArrayList<Item>();
				Item d = getFirstFreeLineEnd();
				// anchor the points
				anchor(FreeItems.getInstance());
				FreeItems.getInstance().clear();
				updateCursor();
				// pick up a copy of all enclosed items
				Collection<Item> enclosedItems = FrameUtils
						.getItemsEnclosedBy(DisplayIO.getCurrentFrame(), d
								.getEnclosedShape());
				if (enclosedItems != null) {
					enclosedItems.removeAll(d.getAllConnected());
					Collection<Item> toCopy = getFullyEnclosedItems(enclosedItems);

					if (toCopy.size() > 0) {
						// Find the closest item to the mouse cursor
						double currentX = DisplayIO.getMouseX();
						double currentY = FrameMouseActions.getY();
						Item closest = null;
						double shortestDistance = Double.MAX_VALUE;
						for (Item next : toCopy) {
							if (next instanceof Line)
								continue;
							double distance = Point.distance((int) currentX, (int) currentY, next.getX(), next.getY());
							if (distance < shortestDistance) {
								shortestDistance = distance;
								closest = next;
							}
						}
						// Move the cursor to closest item
						DisplayIO.setCursorPosition(closest.getPosition());
						// Pickup copy of the stuff inside the rectange
						copies = copy(toCopy);
						pickup(copies);
						// Remove the rectangle
						d.getParentOrCurrentFrame().removeAllItems(
								d.getAllConnected());
					} else {
						// Pick up a copy of the rectangle
						copies = copy(d.getAllConnected());
						pickup(copies);
					}
				}
				return copies;
			}		
		});
		IndirectMouseActions.getInstance().setStampAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				List<Item> copies = new ArrayList<Item>();
				stampItemsOnCursor(true);
				copies = FreeItems.getInstance();
				return copies;
			}	
		});
		IndirectMouseActions.getInstance().setMergeSingleItemAction(new MouseAction() {
			@Override
			public List<Item> exec(final MouseInfo info) {
				List<Item> copies = new ArrayList<Item>();
				copies = copy(FreeItems.getInstance());
				Collection<Item> remain = merge(copies, info.clicked);

				// ignore items that could not be merged.
				anchor(remain);
				return copies;
			}  
		});
		IndirectMouseActions.getInstance().setMergeTwoItemsAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				List<Item> copies = new ArrayList<Item>();
				copies = ItemUtils.UnreelLine(FreeItems.getInstance(),
						_controlDown);
				Collection<Item> leftOver = merge(FreeItems
						.getInstance(), info.clicked);
				anchor(leftOver);
				if (copies == null)
					copies = copy(FreeItems.getInstance());
				FreeItems.getInstance().clear();
				for (Item i : copies)
					i.setOffset(0, 0);
				// need to move to prevent cursor dislocation
				move(copies);
				pickup(copies);
				// point onto point
				return copies;
			} 		
		});
		IndirectMouseActions.getInstance().setMergeGroupAction(new MouseAction() {
			@Override
			public List<Item> exec(final MouseInfo info) {
				List<Item> copies = new ArrayList<Item>();
				// Move the cursor so that the copy is exactly the
				// same as the shape that was anchored
				DisplayIO.setCursorPosition(info.clicked.getPosition());
				Item d = getFirstFreeLineEnd();
				// get a copy of all enclosed items before merging
				// lineEnds
				Collection<Item> items = FrameUtils.getItemsEnclosedBy(
						DisplayIO.getCurrentFrame(), d
								.getEnclosedShape());
				// If its not an enclosed shape then pick up the
				// connected shape
				if (items == null || items.size() == 0) {
					items = d.getAllConnected();
				} else {
					// For some reason the item that was clicked ends up
					// in the enclosure and needs to be removed
					items.removeAll(info.clicked.getConnected());
					// the item that was the origin of the enclosed
					// shape used to create the enclosure does not get
					// returned from getItemsEnclosedBy to the enclosure
					// so it must be added
					items.addAll(d.getConnected());
				}

				Collection<Item> toCopy = new LinkedHashSet<Item>();

				for (Item ip : items) {
					if (ip.hasPermission(UserAppliedPermission.copy))
						toCopy.add(ip);
				}
				copies = copy(toCopy);
				// Now do the merging
				Collection<Item> remain = merge(
						FreeItems.getInstance(), info.clicked);
				// anchor the points
				anchor(remain);
				FreeItems.getInstance().clear();
				pickup(copies);
				return copies;
			}		
		});
		IndirectMouseActions.getInstance().setNewLineAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				List<Item> copies = new ArrayList<Item>();
				Item on = FrameUtils.onItem(DisplayIO.getCurrentFrame(), Math.round(MouseX), Math.round(MouseY), true);
				// If we have permission to copy this item then pick it up
				if (on != null && on.isLineEnd()
						&& on.hasPermission(UserAppliedPermission.full)) {
					on.removeAllConstraints();
					pickup(on);
					return copies;
				}

/*				if (on instanceof WidgetEdge) { TODO: Reinstate. cts16
					// Don't allow the user to break widget edges.
					// Note: had to return here because random dots would
					// appear otherwise... cannot understand code below
					// with create line.
					return copies;
				}*/

				// if its on a line then split the line and put a point on it and
				// pick that point up. Only if it is not a widget line
				if (on instanceof Line && on.hasPermission(UserAppliedPermission.full)) {
					Frame current = DisplayIO.getCurrentFrame();
					// create the two endpoints
					Line oldLine = (Line) info.clicked;
					Item newPoint = oldLine.getStartItem().copy();
					newPoint.setPosition(MouseX, MouseY);

					Item end = oldLine.getEndItem();
					// create the Line
					Line newLine = new Line(newPoint, end, current.getNextItemID());
					oldLine.replaceLineEnd(end, newPoint);
					newPoint.removeAllConstraints();
					pickup(newPoint);
					// Update the stats
					Collection<Item> created = new LinkedList<Item>();
					created.add(newPoint);
					created.add(newLine);
					SessionStats.CreatedItems(newLine.getAllConnected());
					return copies;
				}
				Line newLine = createLine();
				SessionStats.CreatedItems(newLine.getAllConnected());
				return copies;
			}
		});
		IndirectMouseActions.getInstance().setGroupPickupAction(new MouseAction() {
			@Override
			public List<Item> exec(final MouseInfo info) {
				List<Item> copies = new ArrayList<Item>();
				ArrayList<Item> toPickup = new ArrayList<Item>(info.clickedIn.size());
				for (Item ip : info.clickedIn) {
					if (ip.hasPermission(UserAppliedPermission.full)) toPickup.add(ip);
				}
				pickup(toPickup);
				// otherwise the user is creating a line
				return copies;
			} 
		});
		IndirectMouseActions.getInstance().setDetachLineAction(new MouseAction() {
			@Override
			public List<Item> exec(final MouseInfo info) {
				List<Item> items = new ArrayList<Item>();
				// BROOK: WIDGET RECTANGLES DONT ALLOW DISCONNECTION
				if (info.clicked instanceof Line/* && !(info.clicked instanceof WidgetEdge) TODO: Reinstate. cts16*/) {
					// Check if within 20% of the end of the line
					Line l = (Line) info.clicked;
					Item toDisconnect = l.getEndPointToDisconnect(_lastMouseClick.getX(), _lastMouseClick.getY());

					if (toDisconnect == null) {
						pickup(info.clicked);
					} else {
						if (toDisconnect.getHighlightMode() == Item.HighlightMode.Normal) {
							DisplayIO.setCursorPosition(toDisconnect.getPosition(),
									false);
							pickup(toDisconnect);
						} else {
							List<Line> lines = toDisconnect.getLines();
							// This is to remove constraints from single lines
							// with constraints...
							// ie. partially deleted rectangles
							if (lines.size() == 1) {
								toDisconnect.removeAllConstraints();

								DisplayIO.setCursorPosition(toDisconnect
										.getPosition(), false);
								// This is to ensure the selected mode will be set
								// to Normal rather than disconnect when the line is
								// anchored
								toDisconnect
										.setHighlightMode(Item.HighlightMode.Normal);
								pickup(toDisconnect);
							} else {
								// If we are then detatch the line and pick up its
								// end point...
								Frame currentFrame = DisplayIO.getCurrentFrame();
								Item newPoint = null;

								// If the point we are disconnecting is text...
								// Then we want to leave the text behind
								// And disconnect a point
								if (toDisconnect instanceof Text) {
									newPoint = new Dot(toDisconnect.getX(),
											toDisconnect.getY(), -1);
									Item.DuplicateItem(toDisconnect, newPoint);
								} else {
									newPoint = toDisconnect.copy();
								}

								currentFrame.addItem(newPoint);
								// remove the current item from the connected
								// list for this item
								l.replaceLineEnd(toDisconnect, newPoint);
								// remove unneeded constrains
								newPoint.removeAllConstraints();

								// Set the new points mode to normal before picking
								// it up so it will be restored correctly when
								// anchored
								newPoint
										.setHighlightMode(Item.HighlightMode.Normal);
								toDisconnect
										.setHighlightMode(Item.HighlightMode.None);
								DisplayIO.setCursorPosition(toDisconnect
										.getPosition(), false);
								pickup(newPoint);
								ItemUtils.EnclosedCheck(toDisconnect
										.getParentOrCurrentFrame().getItems());
							}
						}
					}
				} else {
					if (info.clicked.isLineEnd()) {
						DisplayIO.setCursorPosition(info.clicked.getPosition(), false);
					}
					pickup(info.clicked);
				}
				return items;
			}
		});
		IndirectMouseActions.getInstance().setAnchorFreeItemsAction(new MouseAction() {
			@Override
			public List<Item> exec(MouseInfo info) {
				List<Item> items = new ArrayList<Item>();
				if (info.clickedIn != null && FreeItems.getInstance().size() == 1) {
					Item item = FreeItems.getItemAttachedToCursor();
					if (item instanceof Text) {
						Text text = (Text) item;
						if (AttributeUtils.setAttribute(text, text, 2)) {
							info.clickedIn.removeAll(FrameUtils
									.getEnclosingLineEnds().iterator().next()
									.getAllConnected());
							for (Item i : info.clickedIn) {
								AttributeUtils.setAttribute(i, text);
							}
							FreeItems.getInstance().clear();
						}
					}
				}

				// if a line is being rubber-banded, check for auto
				// straightening
				anchor(FreeItems.getInstance());
				FreeItems.getInstance().clear();
				updateCursor();
				_offX = _offY = 0;
				return items;
			}  
		});
		IndirectMouseActions.getInstance().setDeleteItemsAction(new MouseAction() {
			@Override
			public List<Item> exec(final MouseInfo info) {
				List<Item> items = new ArrayList<Item>();
				// check permissions
				if (!info.clicked.hasPermission(UserAppliedPermission.full)) {
					//Items on the message box have parent == null
					if (info.clicked.getParent() != null) {
						if (!info.clicked.isFrameName()) {
							Item editTarget = info.clicked.getEditTarget();
							if (editTarget != info.clicked
									&& editTarget
											.hasPermission(UserAppliedPermission.full)) {
								info.clicked = editTarget;
							} else {
								MessageBay
										.displayMessage("Insufficient permission");
								return items;
							}
						}
						
					} else /*Its in the message area*/ {
						MessageBay.displayMessage("Insufficient permission");
						return items;
					}
				}
				Item merger = FreeItems.getItemAttachedToCursor();
				assert (merger != null);
				Collection<Item> left = null;
				// when anchoring a line end onto a text line end, holding shift
				// prevents the line ends from being merged
				if (info.isShiftDown) {
					left = FreeItems.getInstance();
				} else {
					left = merge(FreeItems.getInstance(), info.clicked);
				}
				Collection<Item> toDelete = new LinkedList<Item>();
				toDelete.addAll(FreeItems.getInstance());
				toDelete.removeAll(left);
				anchor(left);
				FreeItems.getInstance().clear();
				DisplayIO.getCurrentFrame().removeAllItems(toDelete);
				updateCursor();
				// Make sure the dot goes away when anchoring a line end behind
				// a text line end
				if (info.isShiftDown) {
					refreshHighlights();
				}
				FrameGraphics.requestRefresh(true);
				return items;
				// otherwise, anchor the items
			}		
		});
	}

	private static FrameMouseActions _instance = null;

	public static FrameMouseActions getInstance() {
		if (_instance == null) _instance = new FrameMouseActions();
		
		return _instance;
	}

	private static final int RECTANGLE_CORNERS = 4;

	// TODO say where/how used
	private static final int MOUSE_WHEEL_THRESHOLD = 2;

	private static final int MINIMUM_RANGE_DEPRESS_TIME = 250;

	private static Date _lastMouseClickDate = new Date();

	public static final int LITTLE_MOUSE_PAUSE = 500;

	public static final int ZERO_MOUSE_PAUSE = 0;

	public static final int BIG_MOUSE_PAUSE = 750;

	public static int _alpha = -1;

	/**
	 * The last known mouse X coordinate
	 */
	public static float MouseX;

	/**
	 * The last known mouse Y coordinate. Relative to the top of the
	 * application.
	 */
	public static float MouseY;

	// Distance of mouse cursor from the origin of the item that was picked up
	// The are used in the move method to calculate the distance moved by the
	// cursor
	private static int _offX;

	private static int _offY;

	// Keeps track of mouse button events when a delete occurs
	private static boolean _isDelete = false;

	// Keeps track of mouse button events when the user extracts attributes
	// occurs
	private static boolean _isAttribute = false;

	/**
	 * A flag to indicate that the last mouseUp event was part of a two button
	 * click sequence hence the next mouse up should be ignored*
	 */
	private static boolean _wasDouble = false;

	private static boolean _isNoOp = false;

	private static boolean _extrude = false;

	// keeps track of the item being 'ranged out' if there is one.
	private static Text _lastRanged = null;

	// keeps track of the picture being cropped if there is one
	private static Picture _lastCropped = null;

	public static int getContext() {
		return _context;
	}

	static int _mouseDown = 0;

	private static MouseEvent _lastMouseClick = null;

	private static Item _lastClickedOn = null;

	private static Collection<Item> _lastClickedIn = null;

	private static boolean _pulseOn = false;

	private static final int PULSE_AMOUNT = 2;

	private static Timer _MouseTimer = new Timer(LITTLE_MOUSE_PAUSE,
			new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					// check if we are in free space
					if (_lastClickedOn == null && FreeItems.getInstance().size() == 0) {
						// System.out.println("SuperBack!");
						_MouseTimer.setDelay(ZERO_MOUSE_PAUSE);
						back();
					} else {
						if (FrameUtils.getCurrentItem() == null) {
							// Check if we are toggling arrowhead
							if (FreeItems.getInstance().size() <= 2) {
								for (Item i : FreeItems.getInstance()) {
									if (i instanceof Line) {
										((Line) i).toggleArrow();
									}
								}
								FrameGraphics.Repaint();
							}
						}
						_MouseTimer.stop();
					}
				}
			});

	private static void setPulse(boolean pulseOn)
	{
		if (_pulseOn == pulseOn) return;

		int amount = PULSE_AMOUNT;
		if (!pulseOn) amount *= -1;
		_pulseOn = pulseOn;

		
		if (_lastClickedOn != null) {
			for (Item i : _lastClickedOn.getAllConnected()) {
				if (i instanceof Line) {
					Line line = (Line) i;
					line.setThickness(line.getThickness() + amount);
				}
			}
		}
		FrameGraphics.Repaint();
	}

	private static Timer _ExtrudeMouseTimer = new Timer(BIG_MOUSE_PAUSE,
			new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					setPulse(true);
					_extrude = true;
					_ExtrudeMouseTimer.stop();
				}
			});

	/**
	 * Each Item on the Frame is checked to determine if the mouse x,y
	 * coordinates are on the Item (or within the Shape surrounding it). If the
	 * coordinates are on the Item then the Item is checked for a link, if it
	 * has a link the link is followed, if not, nothing is done.
	 */
	public void mousePressed(MouseEvent e) {
		ProccessMousePressedEvent(e, e.getModifiersEx());
	}

	private void ProccessMousePressedEvent(MouseEvent e, int modifiersEx) {
		// System.out.println("MousePressed " + e.getX() + "," + e.getY() + " " + e.getWhen());

		// TODO WHY DID I NOT COMMENT THIS LINE!! MIKE SAYS
/*		if (LastRobotX != null) {
			_RobotTimer.stop();
			LastRobotX = null;
			LastRobotY = null;
			mouseMoved(e);
		}
		
		if(ExperimentalFeatures.MousePan.get()) {
    		// don't pan if we're not over the frame
    		_overFrame = FrameUtils.getCurrentItem() == null;
    		_isPanOp = false;
    		// update panning position values so position doesn't jump
    		panStartX = e.getX();
    		panStartY = e.getY();
    		MouseX = panStartX;
    		MouseY = panStartY;
		}*/

		// System.out.println(modifiersEx);
		if (_mouseDown == 0) _lastMouseClickDate = new Date();

		int buttonPressed = e.getButton();
		_mouseDown += buttonPressed;
		//when border of bin is clicked we can pick up the widget has been clicked
		_lastClickedOn = FrameUtils.getCurrentItem();
		// load any frame if necessary
		Item on = _lastClickedOn;

		_lastClickedIn = FrameUtils.getCurrentItems(on);
		// if (_lastClickedIn != null){
		// System.out.println(_lastClickedIn.size());}

		/*
		 * This makes it so clicking repeatedly on the frameName doesn't add the
		 * frames to the backup stack. Only the first frame is added to the
		 * backup stack.
		 */
		if (on == null || buttonPressed != MouseEvent.BUTTON1 || !on.isFrameName()) {
			Navigation.ResetLastAddToBack();
		}

		//SessionStats.MouseClicked(e.getButton());
		if (buttonPressed == MouseEvent.BUTTON1) {
			//SessionStats.AddFrameEvent("Ld");
			_extrude = false;
		} else if (buttonPressed == MouseEvent.BUTTON2) {
			//SessionStats.AddFrameEvent("Md");
			_extrude = false;
		} else if (buttonPressed == MouseEvent.BUTTON3) {
			//SessionStats.AddFrameEvent("Rd");

			// Check if the user picked up a paint brush
			if (FreeItems.getInstance().size() == 1 && FreeItems.getItemAttachedToCursor().isAutoStamp()) {
				int delay = (int) (FreeItems.getItemAttachedToCursor().getAutoStamp() * 1000);
				if (delay < 10) {
					_autoStamp = true;
				} else {
					_autoStampTimer.setDelay(delay);
					_autoStampTimer.start();
				}
			}
		}

		// Mike says...
		// For some reason the modifiers for e are different from modifiersEx
		// The SwingUtilities.convertMouseEvent method changes the modifiers
		_lastMouseClick = e;
		_lastMouseClickModifiers = modifiersEx;
		
		/*
		 * Only start the timer when in free space when the user double clicks
		 * to do super back TODO change this so that there are separate timers
		 * for super back and the other Long depress actions if that is what is
		 * wanted.
		 */
		if (_lastClickedOn == null && FreeItems.getInstance().size() == 0) {
			if (e.getClickCount() >= 2) {
				_MouseTimer.start();
			}
		} else if (	_lastClickedOn != null &&
					FreeItems.getInstance().size() == 0 &&
					e.getButton() == MouseEvent.BUTTON3)
		{
			_ExtrudeMouseTimer.start();
		} else {
			_MouseTimer.start();
		}

		// pre-cache the frame if it is linked

		// If pre-caching is done, it must be done in the background

		// if (on != null && on.getLink() != null && on.isLinkValid()) {
		// FrameIO.Precache(on.getAbsoluteLink());
		// }

		// check for delete command
		if (isDelete(modifiersEx)) {
			_isDelete = true;
			// _lastRanged = null;
			_lastCropped = null;
			_wasDouble = false;
			// check for attributes command
		} else if (isGetAttributes(modifiersEx)) {
			_isAttribute = true;
			_wasDouble = false;
		} else if (isTwoClickNoOp(modifiersEx)) {
			_isAttribute = false;
			_wasDouble = false;
			_isDelete = false;
			_isNoOp = true;
		} else {
			_isDelete = false;
		}

		// This must happen before the previous code
		// This is when the user is anchoring something
		if (	buttonPressed != MouseEvent.BUTTON1 &&
				(_context == CONTEXT_FREESPACE || _context == CONTEXT_AT_ENCLOSURE) &&
				FreeItems.hasItemsAttachedToCursor())
		{
			FrameGraphics.changeHighlightMode(_lastHighlightedItem, Item.HighlightMode.None);

			_lastHighlightedItem = FreeItems.getItemAttachedToCursor();
			for (Item i : FreeItems.getInstance()) {
				i.setHighlightColor(Item.DEPRESSED_HIGHLIGHT);
			}
			FrameGraphics.Repaint();
		// this is when the user is picking something up
		} else if (_lastHighlightedItem != null) {
			if (!(_lastHighlightedItem instanceof Line)) {
				_lastHighlightedItem.setHighlightColor(Item.DEPRESSED_HIGHLIGHT);
			} else {
				for (Item i : _lastHighlightedItem.getAllConnected()) {
					i.setHighlightColor(Item.DEPRESSED_HIGHLIGHT);
				}
			}
			FrameGraphics.Repaint();
		}

		// if the user is ranging text
		if (on != null && on instanceof Text && !_isDelete) {
			_lastRanged = (Text) on;
			// set start-drag point
			_lastRanged.setSelectionStart(DisplayIO.getMouseX(), DisplayIO.getMouseY());
		}

		/*
		 * Want to repaint the text with deleteRange color as soon as the second
		 * button is pressed
		 */
		if (_lastRanged != null) {
			_lastRanged.invalidateAll();
			FrameGraphics.requestRefresh(true);
		}

		if (on != null && on instanceof Picture && e.getButton() == MouseEvent.BUTTON3 && !_isDelete) {
			_lastCropped = (Picture) on;
			// set start crop point
			_lastCropped.setStartCrop(DisplayIO.getMouseX(), DisplayIO.getMouseY());
			_lastCropped.setShowCrop(true);
		}
	}

	// This is where all the processing happens
	public void mouseReleased(MouseEvent e) {

		// System.out.println("Released " + e.getX() + "," + e.getY() + " " +
		// e.getWhen());
//		FrameUtils.ResponseTimer.restart();
		_autoStampTimer.stop();
		_autoStamp = false;

		// Auto-hide popups when user clicks into expeditee world
		// If the user clicks into empty space and a popup-is showing, then
		// the user probably wants to click away the popup - therefore ignore
		// the event
/*		boolean shouldConsume = PopupManager.getInstance().shouldConsumeBackClick(); TODO: Reinstate. cts16
		PopupManager.getInstance().hideAutohidePopups();
		if (shouldConsume && e.getButton() == MouseEvent.BUTTON1) {
			return; // consume back click event
		}*/

		// _lastMovedDistance = new Point(e.getX() - _lastMouseClick.getX(), e
		// .getY()
		// - _lastMouseClick.getY());

		_mouseDown -= e.getButton();
		updateCursor();

		// System.out.println(e.getX() + ", " + e.getY());

		Text lastRanged = _lastRanged;
		_lastRanged = null;
		// Dont do ranging if the user moves really quickly...
		// They are probably trying to pick something up in this case
		if (lastRanged != null) {
			long depressTime = (new Date()).getTime() - _lastMouseClickDate.getTime();
			// double changeInDistance =
			// e.getPoint().distance(_currentMouseClick.getPoint());
			// double speed = changeInDistance * 1000 / changeInTime;

			// System.out.println(depressTime);

			if (depressTime < MINIMUM_RANGE_DEPRESS_TIME || lastRanged.getSelectionSize() <= 0) {// Text.MINIMUM_RANGED_CHARS)
				// {
				lastRanged.clearSelection();
				lastRanged = null;
			}
		}

		_ExtrudeMouseTimer.stop();
		_MouseTimer.stop();
		
		setPulse(false);

		// if the last action was a delete, then ignore the next mouseup
		if (_wasDouble) {
			_wasDouble = false;
			return;
		}

		// This code must come after the _wasDouble code...
		// Otherwise get Stopping Agent method after doing the left+right format
		// shortcut
		if (Actions.isAgentRunning()) {
			Actions.stopAgent();
			return;
		}

		/*
		 * if (_isNoOp) { if (e.getButton() != MouseEvent.NOBUTTON) { _isNoOp =
		 * false; _wasDouble = true; // lastRanged.clearSelection();
		 * FrameGraphics.Repaint(); return; } }
		 */

		// get whatever the user was pointing at
		Item clickedOn = _lastClickedOn;
		Collection<Item> clickedIn = _lastClickedIn;

		MouseX = e.getX();
		MouseY = e.getY();

		Item releasedOn = FrameUtils.getCurrentItem();
		Collection<Item> releasedIn = FrameUtils.getCurrentItems(releasedOn);

		// Only a no op if user releases in free space!
		if (_isPanOp || (_isNoOp && (releasedOn == null && releasedIn == null))) {
			if (_isDelete) {
				_isDelete = false;
				_wasDouble = true;
			}

			_isNoOp = false;

			if (_lastHighlightedItem != null)
				FrameGraphics.changeHighlightMode(_lastHighlightedItem, Item.HighlightMode.None);

			if (FreeItems.hasItemsAttachedToCursor()) {
				move(FreeItems.getInstance());
			}

			if (FreeItems.hasCursor()) {
				move(FreeItems.getCursor(), true);
			}

			if(!_isPanOp) {
				MessageBay.displayMessage("Action cancelled, mouse moved more than "
							+ UserSettings.NoOpThreshold.get() + " pixels.");
			}
			FrameGraphics.Repaint();
			return;
		} else {
			_isNoOp = false;
		}

		// if this is a delete command
		if (_isDelete) {
			if (lastRanged != null) {

				Item i = FreeItems.getItemAttachedToCursor();
				if (i != null && i instanceof Text) {
					lastRanged.replaceSelectedText(((Text) i).getText());
					FreeItems.getInstance().clear();
				} else {
					lastRanged.cutSelectedText();
				}
				lastRanged.clearSelection();
				FrameGraphics.Repaint();

			} else {
				delete(clickedOn);
			}
			_wasDouble = true;
			_isDelete = false;
			return;
		}

		// if this is an attribute extraction command
		if (_isAttribute) {
			if (clickedOn == null) {
				Frame current = DisplayIO.getCurrentFrame();
				if (isControlDown()) {
					Actions.PerformActionCatchErrors(current, null, "HFormat");
				}
				if (!isControlDown() || isShiftDown()) {
					Actions.PerformActionCatchErrors(current, null, "Format");
				}
			} else {
				extractAttributes(clickedOn);
			}
			// if the user dragged and displayed some cropping with left and
			// right button is a no op for now
			// but later could make this the shrinkTo context
			if (_lastCropped != null) {
				_lastCropped.clearCropping();
				_lastCropped = null;
			}
			_wasDouble = true;
			_isAttribute = false;
			return;
		}

		// if the user is ranging-out text
		if (lastRanged != null && e.getButton() != MouseEvent.BUTTON1) {

			Text ranged;
			if (isShiftDown()) {
				// If shift is down, copy everything (size, color, etc.) except actions, links and data
				ranged = lastRanged.copy();
				ranged.setActions(null);
				ranged.setData((List<String>) null);
				ranged.setLink(null);
			} else {
				// If shift isn't down, don't copy any attributes, but base the new text item on the appropriate template
				ranged = DisplayIO.getCurrentFrame().getItemTemplate(lastRanged.copySelectedText().charAt(0));
			}

			// if the user is cutting text from the item
			if (e.getButton() == MouseEvent.BUTTON2) {
				// Check if the user is trying to range an item for which they
				// do not have permission to do so... or it is the frame name
				if (!lastRanged.hasPermission(UserAppliedPermission.full) || lastRanged.isFrameName()) {
					MessageBay.displayMessage("Insufficient permission to cut text");
					lastRanged.clearSelection();
					FrameGraphics.Repaint();
					return;
				}
				// if the entire text is selected and its not a line end then
				// pickup the item
				boolean entireText = lastRanged.getSelectionSize() == lastRanged.getLength();
				if (entireText && !lastRanged.isLineEnd()) {
					lastRanged.clearSelection();
					ranged.delete();
					middleButton(clickedOn, clickedIn, e.isShiftDown());
					return;
				} else {
					ranged.setText(lastRanged.cutSelectedText());
					ranged.setWidth(lastRanged.getWidth());
					// If its the whole text then replace last ranged with a dot
					if (entireText) {
						Item dot = FrameKeyboardActions.replaceText(lastRanged);
						dot.setHighlightMode(HighlightMode.None);
					}
				}
				// if the user is copying text from the item
			} else if (e.getButton() == MouseEvent.BUTTON3) {
				// Check if the user is trying to range an item for which they
				// do not have permission to do so... or it is the frame name
				if (!lastRanged.hasPermission(UserAppliedPermission.copy)) {
					MessageBay.displayMessage("Insufficient permission to copy text");
					lastRanged.clearSelection();
					FrameGraphics.Repaint();
					return;
				}
				ranged.setText(lastRanged.copySelectedText());
			}

			ranged.setParent(null);
			ranged.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
			pickup(ranged);
			lastRanged.clearSelection();
			lastRanged.setHighlightMode(HighlightMode.None);
			refreshHighlights();
			FrameGraphics.refresh(false);
			return;
		}

		// if the user is cropping an image
		if (clickedOn != null && clickedOn == _lastCropped) {
			if (_lastCropped.isCropTooSmall()) {
				_lastCropped = null;
				// FrameGraphics
				// .WarningMessage("Crop cancelled because it was below the
				// minimum size");
			} else {
				Picture cropped = _lastCropped.copy();
				cropped.setParent(null);
				// move the cropped image to the cursor
				int width = cropped.getWidth();
				int height = cropped.getHeight();
				if(cropped.getSource().getX() + width < MouseX) {
					cropped.getSource().setX(MouseX - width);
				}
				if(cropped.getSource().getY() + height < MouseY) {
					cropped.getSource().setY(MouseY - height);
				}
				pickup(cropped);
				// MIKE put the code below up here
				_lastCropped.clearCropping();
				FrameGraphics.changeHighlightMode(_lastCropped, HighlightMode.None);
				_lastCropped = null;
				FrameGraphics.Repaint();
				return;
			}
		}

		assert (_lastCropped == null);
		// if the user has cropped an image, either the above happend or this is
		// a no-op MIKE says WHEN DO WE NEED THE CODE BELOW
		// if (_lastCropped != null && !_lastCropped.isCropTooSmall()) {
		// _lastCropped.clearCropping();
		// _lastCropped = null;
		// FrameGraphics.Repaint();
		// return;
		// }

		// if the user is left-clicking
		if (e.getButton() == MouseEvent.BUTTON1) {
			SessionStats.AddFrameEvent("Lu");
			leftButton(clickedOn, clickedIn, e.isShiftDown(), e.isControlDown());
			return;
		}

		if (e.getButton() == MouseEvent.BUTTON2) {
			SessionStats.AddFrameEvent("Mu");
			middleButton(clickedOn, clickedIn, e.isShiftDown());
			return;
		}

		if (e.getButton() == MouseEvent.BUTTON3) {
			SessionStats.AddFrameEvent("Ru");
			rightButton(clickedOn, clickedIn);
			return;
		}

		// error, we should have returned by now
		System.out.println("Error: mouseReleased should have returned by now. " + e);
	}

	/**
	 * This method handles all left-click actions
	 */
	private void leftButton(Item clicked, Collection<Item> clickedIn, boolean isShiftDown, boolean isControlDown)
	{
		
		//Gets the current frame
		Frame f = DisplayIO.getCurrentFrame();
		
		//Checks if the current frame is an overlay
		if (f.getOverlays() != null && FrameUtils.getCurrentItem() != null) {
			Item i = FrameUtils.getCurrentItem();
			
			//Checks if the item clicked in the overlay is a Rubbish Bin. If it is, delete the item attached to the cursor and return.
/*			if(i instanceof WidgetCorner){ TODO: Reinstate. cts16
				
				try{
					WidgetCorner wc = (WidgetCorner)i;					
					ButtonWidget bw = (ButtonWidget) wc.getWidgetSource();
					
					//Should call a button widgets 'itemheldwhileclicked' method, and process depending on the widget type - else will return false.
					if(bw.itemHeldWhileClicked((InteractiveWidget)bw) == true){									
						
						return;
					}				
				}
				catch (Exception e){
					
					e.printStackTrace();
				}
			}*/
			
			Item on = _lastClickedOn;
			_lastClickedIn = FrameUtils.getCurrentItems(on);
		}
		

		// if the user is pointing at something then either follow the link or
		// do TDFC
		if (clicked == null) {
			// Check if the user is nearby another item...
			int mouseX = DisplayIO.getMouseX();
			int mouseY = DisplayIO.getMouseY();
			// System.out.println(mouseX + "," + mouseY);
			for (Item i : DisplayIO.getCurrentFrame().getItems()) {
				//System.out.println(i.getName().toString());
				if (i instanceof Text) {
					if (i.isNear(mouseX, mouseY)) {
						clicked = i;
						break;
					}
				}
			}
		}

		if (clicked instanceof Text) {
			Text text = (Text) clicked;
			/* Don't follow link when just highlighting text with the left button */
			if (text.getText().length() == 0)
				clicked = null;
			else if (text.getSelectionSize() > 0) {
				return;
			}
		}
		
		// If the user clicked into a widgets free space...
		if (clicked == null && _lastClickedIn != null && _lastClickedIn.size() >= 4) {

			// Check to see if the user clicked into a widgets empty space
/*			InteractiveWidget iw = null; TODO: Reinstate. cts16

			for (Item i : _lastClickedIn) {

				if (i instanceof WidgetCorner) {
					iw = ((WidgetCorner) i).getWidgetSource();
					break;
				} else if (i instanceof WidgetEdge) {
					iw = ((WidgetEdge) i).getWidgetSource();
					break;
				}
			}

			if (iw != null) {

				// Handle dropping items on widgets
				if(iw.ItemsLeftClickDropped()) {
					return;
				}
				
				// Note: mustn't directly use source for handling the link
				// because all link operations will by-pass the widgets special
				// handling with links...
				Item widgetLink = iw.getItems().get(0);
				assert (widgetLink != null);
				clicked = widgetLink;
			} else*/ {
				for (Item i : _lastClickedIn) {
					/*
					 * Find the first linked item or the first unlinked Dot This
					 * code assumes that items are are ordered from top to
					 * bottom. TODO make sure the list will always be ordered
					 * correctly!!
					 */
					if (i.hasLink() || i instanceof Dot) {
						clicked = i;
						break;
					}
				}
			}

		}

		
		if (clicked instanceof Picture) {
			int mouseX = DisplayIO.getMouseX();
			int mouseY = FrameMouseActions.getY();
			Picture clickedOnPicture = (Picture)clicked;
			Frame current_frame = DisplayIO.getCurrentFrame();
			Colour bg_col = current_frame.getBackgroundColor();
			if (clickedOnPicture.MouseOverBackgroundPixel(mouseX,mouseY,bg_col)) {
				// Make 'clicked' null, effectively causing a back() operation
				clicked = null;
			}
		}
		
		if (clicked != null) {
			// check item permissions
			boolean hasLinkOrAction = clicked.hasLink() || clicked.hasAction();

			if ((hasLinkOrAction && !clicked.hasPermission(UserAppliedPermission.followLinks))
					|| (!hasLinkOrAction && !clicked.hasPermission(UserAppliedPermission.createFrames)))
			{
				Item editTarget = clicked.getEditTarget();
				if (editTarget != clicked) {
					if (editTarget.hasPermission(UserAppliedPermission.followLinks)) {
						clicked = editTarget;
					} else {
						MessageBay.displayMessage("Insufficient permission to perform action on item");
						return;
					}
				}
			}

			Item clickedOn = clicked;

			// actions take priority
			if (_lastMouseClick != null && !_lastMouseClick.isControlDown() && clickedOn.hasAction()) {
				IndirectMouseActions.getInstance().getExecuteActionAction().exec(new MouseInfo(clicked, clickedIn, isShiftDown, isControlDown));
			} else if (clickedOn.getLink() != null) {
				IndirectMouseActions.getInstance().getFollowLinkAction().exec(new MouseInfo(clicked, clickedIn, isShiftDown, isControlDown));
			// no link is found, perform TDFC
			} else {
				/*
				 * if the user is clicking on the frame name then move to the
				 * next or previous frame regardless of whether or not the frame
				 * is protected
				 */
				if (clickedOn.isFrameName()) {
					IndirectMouseActions.getInstance().getRespondToFrameNameClickAction().exec(new MouseInfo(clicked, clickedIn, isShiftDown, isControlDown));
				}
				
				IndirectMouseActions.getInstance().getTDFCAction().exec(new MouseInfo(clicked, clickedIn, isShiftDown, isControlDown));
			}

		} else {
			
			IndirectMouseActions.getInstance().getBackAction().exec(new MouseInfo(clicked, clickedIn, isShiftDown, isControlDown));
		}
	}

	public static void middleButton() {
		Item currentItem = FrameUtils.getCurrentItem();
		getInstance().middleButton(currentItem,
				FrameUtils.getCurrentItems(currentItem), false);
		updateCursor();
	}

	public static void rightButton() {
		Item currentItem = FrameUtils.getCurrentItem();
		getInstance().rightButton(currentItem,
				FrameUtils.getCurrentItems(currentItem));
		updateCursor();
	}

	public static void leftButton() {
		Item currentItem = FrameUtils.getCurrentItem();
		getInstance().leftButton(currentItem, FrameUtils.getCurrentItems(currentItem), false, false);
		updateCursor();
	}

	/**
	 * This method handles all middle-click actions
	 */
	private void middleButton(Item clicked, Collection<Item> clickedIn, boolean isShiftDown)
	{
		
		// If the user clicked into a widgets free space...
		if (clicked == null && _lastClickedIn != null
				&& _lastClickedIn.size() >= 4) {

			// Check to see if the use clicked into a widgets empty space
/*			InteractiveWidget iw = null; TODO: Reinstate. cts16

			for (Item i : _lastClickedIn) {

				if (i instanceof WidgetCorner) {
					iw = ((WidgetCorner) i).getWidgetSource();
					break;
				} else if (i instanceof WidgetEdge) {
					iw = ((WidgetEdge) i).getWidgetSource();
					break;
				}
			}

			if (iw != null) {

				// Handle dropping items on widgets
				if(iw.ItemsMiddleClickDropped()) {
					return;
				}
			}*/
		}
		// if the cursor has Items attached
		if (FreeItems.hasItemsAttachedToCursor()) {
			// if the user is pointing at something, merge the items (if possible)
			if (doMerging(clicked)) {
				IndirectMouseActions.getInstance().getDeleteItemsAction().exec(new MouseInfo(clicked, clickedIn, false, false));
			} else {
				IndirectMouseActions.getInstance().getAnchorFreeItemsAction().exec(new MouseInfo(clicked, clickedIn, false, false));
			}
			// otherwise if the user is pointing at something, pick it up unless shift is down
		} else {
			if (clicked instanceof Picture) {
				int mouseX = DisplayIO.getMouseX();
				int mouseY = FrameMouseActions.getY();
				Picture clickedOnPicture = (Picture)clicked;
				Frame current_frame = DisplayIO.getCurrentFrame();
				Colour bg_col = current_frame.getBackgroundColor();
				if (clickedOnPicture.MouseOverBackgroundPixel(mouseX,mouseY,bg_col)) {
					clicked = null; // Effectively make it as if they haven't clicked on anything
				}
			}
			if (clicked != null && !isShiftDown) {

				// check permissions
				if (!clicked.hasPermission(UserAppliedPermission.full)) {
					Item editTarget = clicked.getEditTarget();
					if (editTarget != clicked && editTarget.hasPermission(UserAppliedPermission.full)) {
						clicked = editTarget;
					} else {
						MessageBay.displayMessage("Insufficient permission to pick up item");
						return;
					}
				}
				IndirectMouseActions.getInstance().getDetachLineAction().exec(new MouseInfo(clicked, clickedIn, false, false));
				// if we're inside a shape, pick it up unless shift is down
			} else if (clickedIn != null && !isShiftDown) {
				IndirectMouseActions.getInstance().getGroupPickupAction().exec(new MouseInfo(clicked, clickedIn, false, false));
			} else {
				IndirectMouseActions.getInstance().getNewLineAction().exec(new MouseInfo(clicked, clickedIn, false, false));
			}
		}
		SessionStats.MovedItems(FreeItems.getInstance());
	}

	/**
	 * This method handles all right-click action
	 */
	private void rightButton(Item clicked, Collection<Item> clickedIn) {
		
		// If the user clicked into a widgets free space...
		if (clicked == null && _lastClickedIn != null
				&& _lastClickedIn.size() >= 4) {

			// Check to see if the use clicked into a widgets empty space
/*			InteractiveWidget iw = null; TODO: Reinstate. cts16

			for (Item i : _lastClickedIn) {

				if (i instanceof WidgetCorner) {
					iw = ((WidgetCorner) i).getWidgetSource();
					break;
				} else if (i instanceof WidgetEdge) {
					iw = ((WidgetEdge) i).getWidgetSource();
					break;
				}
			}

			if (iw != null) {

				// Handle dropping items on widgets
				if(iw.ItemsRightClickDropped()) {
					return;
				}
			}*/
		}
		
		// if the cursor has Items attached, then anchor a copy of them

		List<Item> copies = null;
		if (FreeItems.hasItemsAttachedToCursor()) {
			if (FreeItems.getInstance().size() == 1 && FreeItems.getItemAttachedToCursor().isAutoStamp()) {
				// Dont stamp if the user is painting... because we dont want to
				// save any of the items created!
				return;
			// if the user is clicking on something, merge the items unless
			// it is a point onto something other than a lineEnd or a dot
			} else if (clicked != null
			// TODO Change the items merge methods so the logic is simplified
					&& (!(FreeItems.getItemAttachedToCursor() instanceof Dot)
							|| clicked instanceof Dot || clicked.isLineEnd())) {
				// check permissions
				if (!clicked.hasPermission(UserAppliedPermission.full) && clicked.getParent().getNameItem() != clicked) {
					MessageBay.displayMessage("Insufficient permission to merge items");
					return;
				}
				if (clicked instanceof Text || clicked instanceof Dot || clicked instanceof XRayable) {
					if (isRubberBandingCorner()) {
						copies = IndirectMouseActions.getInstance().getMergeGroupAction().exec(new MouseInfo(clicked, clickedIn, false, false));
						// line onto something
					} else if (FreeItems.getInstance().size() == 2 /* && clicked instanceof XRayable */) {
						copies = IndirectMouseActions.getInstance().getMergeTwoItemsAction().exec(new MouseInfo(clicked, clickedIn, false, false));
					} else if (FreeItems.getInstance().size() == 1) {
						copies = IndirectMouseActions.getInstance().getMergeSingleItemAction().exec(new MouseInfo(clicked, clickedIn, false, false));
					} else {
						copies = IndirectMouseActions.getInstance().getStampAction().exec(new MouseInfo(clicked, clickedIn, false, false));
					}
				} else {
					copies = ItemUtils.UnreelLine(FreeItems.getInstance(), _controlDown);
					if (copies == null) copies = copy(FreeItems.getInstance());
					for (Item i : copies) {
						i.setOffset(0, 0);
					}
					anchor(FreeItems.getInstance());
					FreeItems.getInstance().clear();
					pickup(copies);
				}
			// otherwise, anchor the items
			} else {
				// check if this is anchoring a rectangle
				if (isRubberBandingCorner()) {
					copies = IndirectMouseActions.getInstance().getRubberBandingCornerAction().exec(new MouseInfo(clicked, clickedIn, false, false));
				} else {
					if (rubberBanding()) {
						copies = IndirectMouseActions.getInstance().getRubberBandingCopyAction().exec(new MouseInfo(clicked, clickedIn, false, false));
					} else if (_extrude) {
						copies = IndirectMouseActions.getInstance().getExtrudeAction().exec(new MouseInfo(clicked, clickedIn, false, false));
					} else {
						stampItemsOnCursor(true);
						copies = FreeItems.getInstance();
					}
				}
			}
		} else {
			
			if (clicked instanceof Picture) {
				int mouseX = DisplayIO.getMouseX();
				int mouseY = FrameMouseActions.getY();
				Picture clickedOnPicture = (Picture)clicked;
				Frame current_frame = DisplayIO.getCurrentFrame();
				Colour bg_col = current_frame.getBackgroundColor();
				if (clickedOnPicture.MouseOverBackgroundPixel(mouseX,mouseY,bg_col)) {
					clicked = null; // Effectively make it as if they haven't clicked on anything
				}
			}
			
			// if the user is pointing at something and shift isn't down, make a copy
			if (clicked != null && !isShiftDown()) {
				// check permissions
				if (clicked.isLineEnd()) {
					if (!clicked.hasPermission(UserAppliedPermission.full)) {
						MessageBay.displayMessage("Insufficient permission to unreel");
						return;
					}
				} else if (!clicked.hasPermission(UserAppliedPermission.copy)) {
					Item editTarget = clicked.getEditTarget();
					if (editTarget != clicked
							&& editTarget.hasPermission(UserAppliedPermission.copy)) {
						clicked = editTarget;
					} else {
						MessageBay.displayMessage("Insufficient permission to copy");
						return;
					}
				}
				
				copies = IndirectMouseActions.getInstance().getMakeCopyAction().exec(new MouseInfo(clicked, clickedIn, false, false));
			} else {
				// if user is pointing in a closed shape and shift isn't down, make a copy of the items inside
				if (clickedIn != null && !isShiftDown()) {
					IndirectMouseActions.getInstance().getMakeGroupCopyAction().exec(new MouseInfo(clicked, clickedIn, false, false));
					// otherwise, create a rectangle
				} else {
					Item on = FrameUtils.onItem(DisplayIO.getCurrentFrame(), MouseX, MouseY, true);
					// if its on a line then create a line from that line
					if (on instanceof Line && on.hasPermission(UserAppliedPermission.full)) {
						IndirectMouseActions.getInstance().getExtendLineAction().exec(new MouseInfo(on, clickedIn, false, false));
					}
					copies = IndirectMouseActions.getInstance().getCreateRectangleAction().exec(new MouseInfo(clicked, clickedIn, false, false));					
				}
			}
		}
		getInstance().refreshHighlights();
		SessionStats.CopiedItems(copies);
		updateCursor();
		FrameGraphics.Repaint();
	}

	/**
	 * @param enclosedItems
	 * @return
	 */
	private static Collection<Item> getFullyEnclosedItems(
			Collection<Item> enclosure) {
		// copy the enclosedItems because the list will be modified
		Collection<Item> enclosedItems = new LinkedHashSet<Item>(enclosure);
		Collection<Item> toCopy = new LinkedHashSet<Item>(enclosedItems.size());

		while (enclosedItems.size() > 0) {
			Item i = enclosedItems.iterator().next();
			if (i.hasPermission(UserAppliedPermission.copy)) {
				Collection<Item> items = i.getAllConnected();
				// Only copy if the entire shape is enclosed
				if (enclosedItems.containsAll(items)) {
					toCopy.addAll(items);
				}
				enclosedItems.removeAll(items);
			} else {
				enclosedItems.remove(i);
			}
		}
		return toCopy;
	}

//	public void mouseExited(MouseEvent e) {
//	}

	private boolean _overFrame;
	private int panStartX, panStartY;
	private boolean _isPanOp;
	
	public void mouseDragged(MouseEvent e) {
		_lastMouseDragged = e;
		//System.out.println("MouseDragged");

		// Stop the longDepress mouse timer if the user drags above a threshold
		if (_MouseTimer.isRunning()) {
			if (Math.abs(e.getX() - _lastMouseClick.getX()) + Math.abs(e.getY() - _lastMouseClick.getY()) > 10) {
				_MouseTimer.stop();
			}
		}

		if (_autoStamp) {
			stampItemsOnCursor(false);
		}

		// Have the free items follow the cursor if the user clicks in freespace then moves.
		if (FreeItems.getInstance().size() > 0 && _lastClickedOn == null) {
			mouseMoved(e);
			return;
		}
		
		// panning the frame when dragging the mouse while shift-leftclicking
/*		if(ExperimentalFeatures.MousePan.get() && _overFrame && e.isShiftDown() &&
				(e.getModifiersEx() & MouseEvent.BUTTON1_DOWN_MASK) != 0 &&
				(_isPanOp || (Math.max(Math.abs(panStartX - e.getX()), Math.abs(panStartY - e.getY())) > 5)))
		{
			int dX = (int) (e.getX() - MouseX);
			int dY = (int) (e.getY() - MouseY);
			Misc.pan(DisplayIO.getCurrentFrame(), dX, dY);
			MouseX = e.getX();
			MouseY = e.getY();
			_isPanOp = true;
		}*/

		// check if user is dragging across a text item
		if (_lastRanged != null) {
			// System.out.println(MouseY - e.getY());

			MouseX = e.getX();
			MouseY = e.getY();

			int distance = _lastRanged.getY() - FrameMouseActions.getY();
			if (distance <= 0) distance = FrameMouseActions.getY() - _lastRanged.getY() - _lastRanged.getBoundsHeight();

			if (distance > UserSettings.NoOpThreshold.get()) {
				_lastRanged.clearSelectionEnd();
				_isNoOp = true;
			} else {
				// update the ranged section
				_lastRanged.setSelectionEnd(DisplayIO.getMouseX(), DisplayIO.getMouseY());
				_isNoOp = false;
			}

			DisplayIO.setTextCursor(_lastRanged, Text.NONE, false, e.isShiftDown(), e.isControlDown(), false);
			FrameGraphics.Repaint();
			return;
		}

		// if the user is dragging across a picture
		if (_lastCropped != null) {
			// If shift is down then the distance moved is the same in the x and y
			MouseX = e.getX();
			MouseY = e.getY();

			if (e.isControlDown()) {
				int deltaX = Math.abs(e.getX() - _lastMouseClick.getX());
				int deltaY = Math.abs(e.getY() - _lastMouseClick.getY());
				if (deltaX > deltaY) {
					MouseY = _lastMouseClick.getY() + deltaX * (e.getY() > _lastMouseClick.getY() ? 1 : -1);
				} else {
					MouseX = _lastMouseClick.getX() + deltaY * (e.getX() > _lastMouseClick.getX() ? 1 : -1);
				}
			}
			// update the ranged section
			_lastCropped.setEndCrop(DisplayIO.getMouseX(), DisplayIO.getMouseY());

			FrameGraphics.Repaint();
			return;
		}

		// This is the context of a user clicking in freespace an dragging onto
		// the edge of a line
		if (	(_mouseDown == MouseEvent.BUTTON2 || _mouseDown == MouseEvent.BUTTON3) &&
				_lastClickedOn == null &&
				_lastClickedIn == null)
		{
			Item on = FrameUtils.onItem(DisplayIO.getCurrentFrame(), e.getX(), e.getY(), true);

			if (FreeItems.getInstance().size() == 0) {
				// if the user can spot-weld, show the virtual spot
				if (on instanceof Line) {
					Line line = (Line) on;
					line.showVirtualSpot(e.getX(), e.getY());
				}
				if (on != null && on.isLineEnd()) {
					_lastHighlightedItem = on;
					on.setHighlightMode(Item.HighlightMode.Normal);
				} else if (_lastHighlightedItem != null) {
					_lastHighlightedItem.setHighlightMode(Item.HighlightMode.None);
					_lastHighlightedItem = null;
				}
			}
		}

		// Use the below calculation for better speed. If it causes problems
		// switch back to the Euclidean distance calculation
		if (	Math.abs(MouseX - e.getX()) > UserSettings.NoOpThreshold.get() ||
				Math.abs(MouseY - e.getY()) > UserSettings.NoOpThreshold.get())
		{
			_isNoOp = true;
		}
		
		FrameGraphics.Repaint();
	}

	private static MouseEvent _lastMouseMoved = null;

	private static Timer _autoStampTimer = new Timer(200, new ActionListener() {
		public void actionPerformed(ActionEvent ae) {
			stampItemsOnCursor(false);
		}
	});

	private static boolean _controlDown;

	private static boolean _shiftDown;
	
	private static OnNewFrameAction _onFrameAction = null;
	
	public static void setTDFCAction(OnNewFrameAction action) {
	    _onFrameAction = action;
	}

	/**
	 * Updates the stored mouse position and highlights any items as necessary.
	 */
	public void mouseMoved(MouseEvent e) {
		mouseMoved(e, false);
	}

	

	


	

	

	

	/*
	 * private boolean mouseMovedRecently() { Date now = new Date();
	 * 
	 * return now.getTime() - _lastMouseMovement.getTime() < 150; }
	 */

	public void mouseWheelMoved(MouseWheelEvent arg0) {
		Navigation.ResetLastAddToBack();

		int clicks = Math.abs(arg0.getWheelRotation());

		if (FreeItems.getInstance().size() == 2) {
			if ((FreeItems.getInstance().get(0).isLineEnd() && FreeItems.getInstance().get(1) instanceof Line) ||
					(FreeItems.getInstance().get(1).isLineEnd() && FreeItems.getInstance().get(0) instanceof Line))
			{
				Line line;
				if (FreeItems.getInstance().get(0) instanceof Line)
					line = (Line) FreeItems.getInstance().get(0);
				else
					line = (Line) FreeItems.getInstance().get(1);

				// User must do multiple clicks to toggle the line
				if (clicks >= MOUSE_WHEEL_THRESHOLD) {
					if (arg0.isShiftDown())
						line.toggleArrowHeadRatio(arg0.getWheelRotation());
					else
						line.toggleArrowHeadLength(arg0.getWheelRotation());
					// line.getParent().change();
					FrameGraphics.Repaint();
				}
			}
		} else if (arg0.getWheelRotation() != 0 && arg0.isShiftDown()) {

			FunctionKey rotationType = FunctionKey.SizeUp;
			if (arg0.getWheelRotation() > 0) {
				rotationType = FunctionKey.SizeDown;
			}

			FrameKeyboardActions.functionKey(rotationType, 1, arg0.isShiftDown(), arg0.isControlDown());

		} else if (clicks >= MOUSE_WHEEL_THRESHOLD) {
			Item item = FrameUtils.getCurrentItem();

			// if the user is not pointing to any item
			if (item == null) {
				FrameKeyboardActions.NextTextItem(null,
						arg0.getWheelRotation() > 0);
				return;
			}

			if (item instanceof Line || item instanceof Circle) {
				// check permissions
				if (!item.hasPermission(UserAppliedPermission.full)) {
					MessageBay
							.displayMessage("Insufficient permission to edit the Line");
					return;
				}
				item.toggleDashed(arg0.getWheelRotation());
				item.getParent().change();
				FrameGraphics.Repaint();
				return;
			} else if (item instanceof Text) {
				FrameKeyboardActions.NextTextItem(item,
						arg0.getWheelRotation() > 0);
			}
		}
	}

	/**
	 * 
	 * @return the integer value for the last mouse button clicked.
	 */
	public static int getLastMouseButton() {
		if (_lastMouseClick == null) return MouseEvent.NOBUTTON;

		return _lastMouseClick.getButton();
	}

	public static boolean isDelete(int modifiersEx) {
		int onMask = MouseEvent.BUTTON3_DOWN_MASK | MouseEvent.BUTTON2_DOWN_MASK;
		return (modifiersEx & onMask) == onMask;
	}

	public static boolean isGetAttributes(int modifiersEx) {
		int onMask = MouseEvent.BUTTON3_DOWN_MASK | MouseEvent.BUTTON1_DOWN_MASK;
		return (modifiersEx & onMask) == onMask;
	}

	public static boolean isTwoClickNoOp(int modifiersEx) {
		int onMask = MouseEvent.BUTTON2_DOWN_MASK | MouseEvent.BUTTON1_DOWN_MASK;
		return (modifiersEx & onMask) == onMask;
	}

	public static boolean wasDeleteClicked() {
		if (_lastMouseClick == null)
			return false;
		return isDelete(_lastMouseClickModifiers);
	}

	public static void control(KeyEvent ke) {
		for (Item i : FreeItems.getInstance()) {
			i.invalidateCommonTrait(ItemAppearence.PreMoved);
		}

		for (Item i : FreeItems.getCursor()) {
			i.invalidateCommonTrait(ItemAppearence.PreMoved);
		}

		_controlDown = ke.isControlDown();

		if (_controlDown) {
			// TODO why are these two lines needed?!?!
			// _offX = 0;
			// _offY = 0;
		} else {
			resetOffset();
		}

		if (_mouseDown > 0 && _lastMouseDragged != null) {
			MouseEvent me = _lastMouseDragged;
			_lastMouseDragged = new MouseEvent(ke.getComponent(),
					MouseEvent.NOBUTTON, ke.getWhen(), ke.getModifiers(), me
							.getX(), me.getY(), 0, false);
			_instance.mouseDragged(_lastMouseDragged);
		} else if (_lastMouseMoved != null) {
			MouseEvent me = _lastMouseMoved;
			_lastMouseMoved = new MouseEvent(ke.getComponent(),
					MouseEvent.NOBUTTON, ke.getWhen(), ke.getModifiers(), me
							.getX(), me.getY(), 0, false);
			_instance.mouseMoved(_lastMouseMoved, true);
		}
		updateCursor();
		
		Help.updateStatus();

		for (Item i : FreeItems.getInstance()) {
			i.invalidateCommonTrait(ItemAppearence.PostMoved);
		}

		for (Item i : FreeItems.getCursor()) {
			i.invalidateCommonTrait(ItemAppearence.PostMoved);
		}
		// TODO: Check why the new constrained line is not repainted immediately
		FrameGraphics.requestRefresh(true);
		FrameGraphics.refresh(true);
	}

	public static int getX() {
		return Math.round(MouseX);
	}

	public static int getY() {
		return Math.round(MouseY);
	}

	public static Point getPosition() {
		return new Point(getX(), getY());
	}

	public static Point getFreeItemsOffset() {
		return new Point(_offX, _offY);
	}

	public static void shift(KeyEvent e) {
		_shiftDown = e.isShiftDown();
		Help.updateStatus();
		getInstance().refreshHighlights();
	}

	@Override
	public void onGesture(Gesture gesture) {
		// TODO Auto-generated method stub
		
	}
}
